###global _:true,angular:true,###
angular.module('uiGmapgoogle-maps.directives.api.utils')
.service('uiGmap_sync', [ ->
    fakePromise: ->
      _cb = undefined
      then: (cb) ->
        _cb = cb
      resolve: ->
        _cb.apply(undefined, arguments)
  ])
.service 'uiGmap_async', [ '$timeout', 'uiGmapPromise', 'uiGmapLogger', '$q','uiGmapDataStructures', 'uiGmapGmapUtil',
($timeout, uiGmapPromise, $log, $q, uiGmapDataStructures, uiGmapGmapUtil) ->
  promiseTypes = uiGmapPromise.promiseTypes
  isInProgress = uiGmapPromise.isInProgress
  promiseStatus = uiGmapPromise.promiseStatus
  ExposedPromise =  uiGmapPromise.ExposedPromise
  SniffedPromise = uiGmapPromise.SniffedPromise

  kickPromise = (sniffedPromise, cancelCb) ->
    #kick a promise off and log some info on it
    promise = sniffedPromise.promise()
    promise.promiseType = sniffedPromise.promiseType
    $log.debug "promiseType: #{promise.promiseType}, state: #{promiseStatus promise.$$state.status}" if promise.$$state
    promise.cancelCb = cancelCb
    promise

  doSkippPromise = (sniffedPromise,lastPromise) ->
    # note this skipp could be specific to polys (but it works for that)
    if sniffedPromise.promiseType == promiseTypes.create and
      lastPromise.promiseType != promiseTypes.delete and lastPromise.promiseType != promiseTypes.init
        $log.debug "lastPromise.promiseType #{lastPromise.promiseType}, newPromiseType: #{sniffedPromise.promiseType}, SKIPPED MUST COME AFTER DELETE ONLY"
        return true
    false

  maybeCancelPromises = (queue, sniffedPromise,lastPromise) ->
#    $log.warn "sniff: promiseType: #{sniffedPromise.promiseType}, lastPromiseType: #{lastPromise.promiseType}"
#    $log.warn "lastPromise.cancelCb #{lastPromise.cancelCb}"
    if sniffedPromise.promiseType == promiseTypes.delete and lastPromise.promiseType != promiseTypes.delete
      if lastPromise.cancelCb? and _.isFunction(lastPromise.cancelCb) and isInProgress(lastPromise)
        $log.debug "promiseType: #{sniffedPromise.promiseType}, CANCELING LAST PROMISE type: #{lastPromise.promiseType}"
        lastPromise.cancelCb('cancel safe')
        #see if we can cancel anything else
        first = queue.peek()
        if first? and isInProgress(first)# and first.promiseType != promiseTypes.delete
          if first.hasOwnProperty("cancelCb") and _.isFunction first.cancelCb
            $log.debug "promiseType: #{first.promiseType}, CANCELING FIRST PROMISE type: #{first.promiseType}"
            first.cancelCb('cancel safe')
          else
            $log.warn 'first promise was not cancelable'

  ###
  From a High Level:
    This is a SniffedPromiseQueueManager (looking to rename) where the queue is existingPiecesObj.existingPieces.
    This is a function and should not be considered a class.
    So it is run to manage the state (cancel, skip, link) as needed.
  Purpose:
  The whole point is to check if there is existing async work going on. If so we wait on it.

  arguments:
  - existingPiecesObj =  Queue<Promises>
  - sniffedPromise = object wrapper holding a function to a pending (function) promise (promise: fnPromise)
  with its intended type.
  - cancelCb = callback which accepts a string, this string is intended to be returned at the end of _async.each iterator

    Where the cancelCb passed msg is 'cancel safe' _async.each will drop out and fall through. Thus canceling the promise
    gracefully without messing up state.

  Synopsis:

   - Promises have been broken down to 4 states create, update,delete (3 main) and init. (Helps boil down problems in ordering)
    where (init) is special to indicate that it is one of the first or to allow a create promise to work beyond being after a delete

   - Every Promise that comes in is enqueued and linked to the last promise in the queue.

   - A promise can be skipped or canceled to save cycles.

  Saved Cycles:
    - Skipped - This will only happen if async work comes in out of order. Where a pending create promise (un-executed) comes in
      after a delete promise.
    - Canceled - Where an incoming promise (un-executed promise) is of type delete and the any lastPromise is not a delete type.


  NOTE:
  - You should not muck with existingPieces as its state is dependent on this functional loop.
  - PromiseQueueManager should not be thought of as a class that has a life expectancy (it has none). It's sole
  purpose is to link, skip, and kill promises. It also manages the promise queue existingPieces.
  ###
  PromiseQueueManager = (existingPiecesObj, sniffedPromise, cancelCb) ->
    unless existingPiecesObj.existingPieces
      #TODO: rename existingPieces to some kind of queue
      existingPiecesObj.existingPieces = new uiGmapDataStructures.Queue()
      existingPiecesObj.existingPieces.enqueue kickPromise(sniffedPromise, cancelCb)
    else
      lastPromise = _.last existingPiecesObj.existingPieces._content

      return if doSkippPromise(sniffedPromise, lastPromise)
      maybeCancelPromises(existingPiecesObj.existingPieces, sniffedPromise, lastPromise)

      newPromise = ExposedPromise lastPromise.finally ->
       kickPromise(sniffedPromise, cancelCb)
      newPromise.cancelCb = cancelCb
      newPromise.promiseType = sniffedPromise.promiseType
      existingPiecesObj.existingPieces.enqueue newPromise
      # finally is important as we don't care how something is canceled
      lastPromise.finally ->
        # keep the queue tight
        existingPiecesObj.existingPieces.dequeue()

  managePromiseQueue = (objectToLock, promiseType, msg = '', cancelCb, fnPromise) ->
    cancelLogger = (msg) ->
      $log.debug "#{msg}: #{msg}"
      if cancelCb? and _.isFunction cancelCb
        cancelCb(msg)
    PromiseQueueManager objectToLock, SniffedPromise(fnPromise, promiseType), cancelLogger


  defaultChunkSize = 80
  errorObject =
    value: null

  #https://github.com/petkaantonov/bluebird/wiki/Optimization-killers
  tryCatch = (fn, ctx, args) ->
    try
      return fn.apply(ctx, args)
    catch e
      errorObject.value = e
      return errorObject

  logTryCatch = (fn, ctx, deferred, args) ->
    result = tryCatch(fn, ctx, args)
    if result == errorObject
      msg = "error within chunking iterator: #{errorObject.value}"
      $log.error msg
      deferred.reject msg
    if result == 'cancel safe'
      # THIS IS MAD IMPORTANT AS THIS IS OUR FALLTHOUGH TO ALLOW
      # _async.each iterator to drop out at a safe point (IE the end of its iterator callback)
      return false
    true


  _getIterateeValue = (collection, array, index) ->
    # logger.debug "collection: #{JSON.stringify collection}"
    # logger.debug "array: #{JSON.stringify array}"
    # logger.debug "index: #{index}"
    _isArray = collection == array
    valOrKey = array[index]
    return  valOrKey if _isArray
    collection[valOrKey]

  _ignoreFields = ['length', 'forEach', 'map']

  getArrayAndKeys = (collection, keys, bailOutCb, cb) ->
    #checks to handle array and object iteration
    if angular.isArray collection
      array = collection
    else
      if keys
        array = keys
      else
        array = []
        for propName, val of collection
          if collection.hasOwnProperty(propName) and !_.includes(_ignoreFields, propName)
            array.push propName

    if !cb? #shifting args if last cb is not defined
      cb = bailOutCb

    if angular.isArray(array) and !array?.length
      return bailOutCb() if cb != bailOutCb
    cb(array, keys)

  ###
    Author: Nicholas McCready & jfriend00
    _async handles things asynchronous-like :), to allow the UI to be free'd to do other things
    Code taken from http://stackoverflow.com/questions/10344498/best-way-to-iterate-over-an-array-without-blocking-the-ui

    The design of any functionality of _async is to be like lodash/underscore and replicate it but call things
    asynchronously underneath. Each should be sufficient for most things to be derived from.

    Optional Asynchronous Chunking via promises.
  ###
  doChunk = (collection, chunkSizeOrDontChunk, pauseMilli, chunkCb, pauseCb, overallD, index, _keys) ->
    getArrayAndKeys collection, _keys, (array, keys) ->
      # logger.debug array
      # logger.debug keys
      if chunkSizeOrDontChunk and chunkSizeOrDontChunk < array.length
        cnt = chunkSizeOrDontChunk
      else
        cnt = array.length

      i = index
      keepGoing = true
      # logger.debug "cnt: #{cnt}"
      while keepGoing and cnt-- and i < (if array then array.length else i + 1)
        # logger.debug "cnt: #{cnt}"
        # process array[index] here
        val = _getIterateeValue(collection, array, i)
        # logger.debug "val: #{val}"
        keepGoing = if angular.isFunction val then true else logTryCatch chunkCb, undefined, overallD, [val, i]
        ++i

      if array
        if keepGoing and i < array.length
          index = i
          if chunkSizeOrDontChunk
            if pauseCb? and _.isFunction pauseCb
              logTryCatch pauseCb, undefined, overallD, []
            $timeout ->
              doChunk collection, chunkSizeOrDontChunk, pauseMilli, chunkCb, pauseCb, overallD, index, keys
            , pauseMilli, false
        else
          overallD.resolve()

  each = (collection, chunk, chunkSizeOrDontChunk = defaultChunkSize, pauseCb, index = 0, pauseMilli = 1, _keys) ->
    ret = undefined
    overallD = uiGmapPromise.defer()
    ret = overallD.promise

    unless pauseMilli
      error = 'pause (delay) must be set from _async!'
      $log.error error
      overallD.reject error
      return ret

    getArrayAndKeys collection, _keys
    , ->
      overallD.resolve()
      return ret
    , (array, keys) ->
      # set this to whatever number of items you can process at once
      doChunk collection, chunkSizeOrDontChunk, pauseMilli, chunk, pauseCb, overallD, index, keys

      return ret

  #copied from underscore but w/ async each above
  map = (collection, iterator, chunkSizeOrDontChunk, pauseCb, index, pauseMilli, _keys) ->
    results = []
    getArrayAndKeys collection, _keys
    , ->
      return uiGmapPromise.resolve(results)
    , (array, keys) ->
      each(collection, (o) ->
        results.push iterator o
      , chunkSizeOrDontChunk, pauseCb, index, pauseMilli, keys)
      .then ->
        results


  each: each
  map: map
  managePromiseQueue: managePromiseQueue
  promiseLock: managePromiseQueue
  defaultChunkSize: defaultChunkSize
  getArrayAndKeys: getArrayAndKeys
  chunkSizeFrom: (fromSize, ret = undefined) ->
    if _.isNumber fromSize
      ret = fromSize
    if uiGmapGmapUtil.isFalse(fromSize) or fromSize == false
      ret = false
    ret
]
